# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from __future__ import annotations

import itertools
import logging
import pickle

from ..i2c_connection_helper import I2C_Connection_Helper
from .address_space_controller import Address_Space_Controller


class Base_Chip:
    newid = itertools.count()

    def __init__(
        self,
        chip_name: str,
        i2c_connection: I2C_Connection_Helper,
        logger: logging.Logger,
        software_version: str = "",
        register_model=None,
        register_decoding=None,
        indexer_info=None,
    ):
        self._id = next(Base_Chip.newid)

        self._chip_name = chip_name
        self._unique_name = chip_name + "_{}".format(self._id)
        self._software_version = software_version

        self._i2c_connection = i2c_connection
        self._logger = logger

        self._address_space = {}
        self._register_model = register_model
        self._register_decoding = register_decoding

        self._enable_readback = True

        self._indexer_vars = {}
        if indexer_info is not None:
            self._build_indexer_vars(indexer_info)

        for address_space in self._register_model:
            # decoding = None
            # if address_space in self._register_decoding:
            #    decoding = self._register_decoding[address_space]
            self._register_address_space(address_space, None, self._register_model[address_space])  # , decoding)

    def __getitem__(self, index):
        address_space_name, block_name, register = index

        block_ref, _ = self._gen_block_ref_from_indexers(
            address_space_name=address_space_name,
            block_name=block_name,
            full_array=False,
        )

        return self._address_space[address_space_name][block_ref, register]

    def __setitem__(self, index, value):
        address_space_name, block_name, register = index

        block_ref, _ = self._gen_block_ref_from_indexers(
            address_space_name=address_space_name,
            block_name=block_name,
            full_array=False,
        )

        self._address_space[address_space_name][block_ref, register] = value

    @property
    def id(self):
        return self._id

    @property
    def enable_readback(self):
        return self._enable_readback

    @enable_readback.setter
    def enable_readback(self, value: bool):
        self._enable_readback = value

    def _build_indexer_vars(self, indexer_info):
        indexer_variables = indexer_info["vars"]
        variables_min = indexer_info["min"]
        variables_max = indexer_info["max"]

        if len(indexer_variables) != len(variables_min) or len(indexer_variables) != len(variables_max):
            raise RuntimeError(f"Lengths of control structures for the indexer of {self._chip_name} do not match")

        for idx in range(len(indexer_variables)):
            variable = indexer_variables[idx]
            minimum = variables_min[idx]
            maximum = variables_max[idx]

            if variable == "block" and minimum is None and maximum is None:
                continue

            if minimum is None and maximum is None:
                continue
            value = minimum
            if value is None:
                value = maximum

            self._indexer_vars[variable] = {"variable": value, "min": minimum, "max": maximum}

    def _register_address_space(self, name: str, address: int, address_space_model):
        if name in self._address_space:
            raise ValueError("An address space with the name '{}' already exists".format(name))

        size = address_space_model["Memory Size"]
        if "Address Bitlength" in address_space_model:
            address_bitlength = address_space_model["Address Bitlength"]
        else:
            address_bitlength = 8
        if "Word Bitlength" in address_space_model:
            word_bitlength = address_space_model["Word Bitlength"]
        else:
            word_bitlength = 8
        if "Address Endianness" in address_space_model:
            address_endianness = address_space_model["Address Endianness"]
        else:
            address_endianness = "big"
        if "Word Endianness" in address_space_model:
            word_endianness = address_space_model["Word Endianness"]
        else:
            word_endianness = "little"
        if "Read Type" in address_space_model:
            read_type = address_space_model["Read Type"]
        else:
            read_type = "Normal"
        if "Write Type" in address_space_model:
            write_type = address_space_model["Write Type"]
        else:
            write_type = "Normal"

        self._address_space[name] = Address_Space_Controller(
            name=name,
            i2c_address=address,
            address_space_size=size,
            logger=self._logger,
            i2c_connection=self._i2c_connection,
            address_bitlength=address_bitlength,
            address_endianness=address_endianness,
            word_bitlength=word_bitlength,
            word_endianness=word_endianness,
            read_type=read_type,
            write_type=write_type,
            address_space_map=address_space_model["Register Blocks"],
        )

    def _gen_block_ref_from_indexers(self, address_space_name: str, block_name: str, full_array: bool):
        block_ref = block_name
        params = {'block': block_name}

        if "Indexer" in self._register_model[address_space_name]["Register Blocks"][block_name] and not full_array:
            indexers = self._register_model[address_space_name]["Register Blocks"][block_name]["Indexer"]["vars"]
            min_vals = self._register_model[address_space_name]["Register Blocks"][block_name]["Indexer"]["min"]
            max_vals = self._register_model[address_space_name]["Register Blocks"][block_name]["Indexer"]["max"]

            block_ref = ""
            params = {}
            for idx in range(len(indexers)):
                indexer = indexers[idx]
                min_val = min_vals[idx]
                max_val = max_vals[idx]

                if block_ref != "":
                    block_ref += ":"

                if indexer == "block" and min_val is None and max_val is None:
                    block_ref += block_name
                    params[indexer] = block_name
                else:
                    val = self._indexer_vars[indexer]['variable']
                    if val == "":
                        val = 0
                    block_ref += f"{val}"
                    params[indexer] = val

        return block_ref, params

    def get_indexer_info(self, name):
        if name not in self._indexer_vars:
            return None, None, None
        return self._indexer_vars[name]['variable'], self._indexer_vars[name]['min'], self._indexer_vars[name]['max']

    # Promote this to a standalone function
    def get_indexer_array(self, indexer_info):
        indexer_variables = indexer_info["vars"]
        variables_min = indexer_info["min"]
        variables_max = indexer_info["max"]

        indexer_array = {}
        tmp = {}
        for idx in range(len(indexer_variables)):
            variable = indexer_variables[idx]
            if variable == "block":
                if len(indexer_array) == 0:
                    tmp[""] = {"arguments": ["{block}"]}
                else:
                    for tag in indexer_array:
                        tmp[tag] = {"arguments": indexer_array[tag]["arguments"] + ["{block}"]}
            else:
                minimum = variables_min[idx]
                maximum = variables_max[idx]

                if minimum is None and maximum is None:
                    continue

                if minimum is None or maximum is None:
                    if minimum is not None:
                        my_range = [minimum]
                    else:
                        my_range = [maximum]
                elif minimum == 0 and maximum == 1:
                    my_range = [False, True]
                else:
                    my_range = range(minimum, maximum)

                for value in my_range:
                    if len(indexer_array) == 0:
                        tmp["{}".format(value)] = {"arguments": ["{}".format(value)]}
                    else:
                        for tag in indexer_array:
                            if tag == "":
                                index = "{}".format(value)
                            else:
                                index = "{}_{}".format(tag, value)
                            tmp[index] = {"arguments": indexer_array[tag]["arguments"] + [value]}

            indexer_array = tmp
            tmp = {}

        return indexer_array

    def set_indexer(self, name, value):
        minVal = self._indexer_vars[name]['min']
        maxVal = self._indexer_vars[name]['max']
        if minVal is not None and maxVal is not None:
            if value < minVal:
                raise RuntimeError(f"The indexer '{name}' should not have a value smaller than {minVal}, tried to set {value}")
            if value >= maxVal:
                raise RuntimeError(f"The indexer '{name}' should not have a value greater than or equal to {minVal}, tried to set {value}")
        self._indexer_vars[name]['variable'] = value

    def read_all(self):
        for address_space in self._address_space:
            self.read_all_address_space(address_space)

    def write_all(self, readback_check: bool = True):
        success = True
        for address_space in self._address_space:
            if not self.write_all_address_space(address_space, readback_check=readback_check):
                success = False

        return success

    def read_all_address_space(self, address_space_name: str, no_message: bool = True):
        if not no_message:
            self._logger.info("Reading full address space: {}".format(address_space_name))
        address_space: Address_Space_Controller = self._address_space[address_space_name]
        address_space.read_all()

    def write_all_address_space(self, address_space_name: str, readback_check: bool = True, no_message: bool = True):
        if not no_message:
            self._logger.info("Writing full address space: {}".format(address_space_name))
        address_space: Address_Space_Controller = self._address_space[address_space_name]
        return address_space.write_all(readback_check=readback_check)

    def read_all_block(self, address_space_name: str, block_name: str, full_array: bool = False, no_message: bool = True):
        block_ref, _ = self._gen_block_ref_from_indexers(
            address_space_name=address_space_name,
            block_name=block_name,
            full_array=full_array,
        )

        if not no_message:
            self._logger.info(f"Reading block {block_ref} from address space {address_space_name} of chip {self._chip_name}")
        address_space: Address_Space_Controller = self._address_space[address_space_name]
        address_space.read_block(block_ref)

    def write_all_block(
        self, address_space_name: str, block_name: str, full_array: bool = False, readback_check: bool = True, no_message: bool = True
    ):
        block_ref, _ = self._gen_block_ref_from_indexers(
            address_space_name=address_space_name,
            block_name=block_name,
            full_array=full_array,
        )

        if not no_message:
            self._logger.info(f"Writing block {block_ref} from address space {address_space_name} of chip {self._chip_name}")
        address_space: Address_Space_Controller = self._address_space[address_space_name]
        return address_space.write_block(block_ref, readback_check=readback_check)

    def read_register(self, address_space_name: str, block_name: str, register: str, no_message: bool = True):
        self._logger.detailed_trace(f'Base_Chip::read_register("{address_space_name}", "{block_name}", "{register}", {no_message})')
        block_ref, _ = self._gen_block_ref_from_indexers(
            address_space_name=address_space_name,
            block_name=block_name,
            full_array=False,
        )
        # self._logger.detailed_trace(f'   Got block_ref={block_ref}')

        if not no_message:
            self._logger.info(
                "Reading register {} from block {} of address space {} of chip {}".format(
                    register, block_ref, address_space_name, self._chip_name
                )
            )
        address_space: Address_Space_Controller = self._address_space[address_space_name]
        address_space.read_register(block_ref, register)

    def write_register(self, address_space_name: str, block_name: str, register: str, readback_check: bool = True, no_message: bool = True):
        self._logger.detailed_trace(
            f'Base_Chip::write_register("{address_space_name}", "{block_name}", "{register}", "{readback_check}", {no_message})'
        )
        block_ref, _ = self._gen_block_ref_from_indexers(
            address_space_name=address_space_name,
            block_name=block_name,
            full_array=False,
        )

        if not no_message:
            self._logger.info(
                "Writing register {} from block {} of address space {} of chip {}".format(
                    register, block_ref, address_space_name, self._chip_name
                )
            )
        address_space: Address_Space_Controller = self._address_space[address_space_name]
        return address_space.write_register(block_ref, register, readback_check=readback_check)

    def read_decoded_value(self, address_space_name: str, block_name: str, decoded_value_name: str, no_message: bool = False):
        value_info = self._register_decoding[address_space_name]['Register Blocks'][block_name][decoded_value_name]

        for position in value_info['position']:
            register = position[0]
            self.read_register(address_space_name, block_name, register, no_message=no_message)

    def write_decoded_value(
        self, address_space_name: str, block_name: str, decoded_value_name: str, write_check: bool = True, no_message: bool = False
    ):
        value_info = self._register_decoding[address_space_name]['Register Blocks'][block_name][decoded_value_name]

        for position in value_info['position']:
            register = position[0]
            self.write_register(address_space_name, block_name, register, write_check, no_message=no_message)

    def get_register_value(self, address_space_name: str, block_name: str, register: str, no_message: bool = True):
        self._logger.detailed_trace(f'Base_Chip::get_register_value("{address_space_name}", "{block_name}", "{register}", {no_message})')
        block_ref, _ = self._gen_block_ref_from_indexers(
            address_space_name=address_space_name,
            block_name=block_name,
            full_array=False,
        )

        if not no_message:
            self._logger.info(
                "Reading register {} from block {} of address space {} of chip {}".format(
                    register, block_ref, address_space_name, self._chip_name
                )
            )
        address_space: Address_Space_Controller = self._address_space[address_space_name]
        address_space.read_register(block_ref, register)

    def save_pickle_file(self, config_file: str, object):
        save_object = {
            'object': object,
            'chip': self._chip_name,
            'version': self._software_version,
        }

        with open(config_file, 'wb') as f:
            pickle.dump(save_object, f)

    def load_pickle_file(self, config_file: str):
        loaded_obj = None
        with open(config_file, 'rb') as f:
            loaded_obj = pickle.load(f)

        if 'chip' not in loaded_obj:
            self._logger.error("The pickle file does not have the correct format")
            return None

        if loaded_obj['chip'] != self._chip_name:
            self._logger.error(
                "Wrong config file type. It was saved for the chip: {}; expected {}".format(loaded_obj['chip'], self._chip_name)
            )
            return None

        # TODO: for the version we should probably implement some sort of semantic versioning
        if loaded_obj['version'] != self._software_version:
            self._logger.error(
                "Wrong config file type. It was saved for a different version of this chip: {}; expected {}".format(
                    loaded_obj['version'], self._software_version
                )
            )
            return None

        return loaded_obj['object']

    def save_config(self, config_file: str):
        info = {}

        for address_space_name in self._address_space:
            address_space: Address_Space_Controller = self._address_space[address_space_name]
            size = address_space._address_space_size

            conf = [None for _ in range(size)]

            for idx in range(size):
                conf[idx] = address_space._memory[idx]

            info[address_space_name] = conf

        self.save_pickle_file(config_file, info)

    def load_config(self, config_file: str):
        info = self.load_pickle_file(config_file)

        for address_space_name in self._address_space:
            address_space: Address_Space_Controller = self._address_space[address_space_name]
            size = address_space._address_space_size

            for idx in range(size):
                address_space._memory[idx] = info[address_space_name][idx]

    def reset_config_to_default(self):
        for name in self._address_space:
            self._address_space[name].reset_to_defaults()



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from __future__ import annotations

import logging
from math import ceil

from ..i2c_connection_helper import I2C_Connection_Helper
from ..i2c_connection_helper import valid_endianness
from ..i2c_connection_helper import valid_read_type
from ..i2c_connection_helper import valid_write_type


class Address_Space_Controller:
    def __init__(
        self,
        name: str,
        i2c_address: int,
        address_space_size: int,
        logger: logging.Logger,
        i2c_connection: I2C_Connection_Helper,
        address_bitlength: int = 8,
        address_endianness: str = 'big',
        word_bitlength: int = 8,
        word_endianness: str = 'big',
        read_type: str = 'Normal',
        write_type: str = 'Normal',
        address_space_map: dict = {},
    ):
        self._name = name
        self._i2c_address = i2c_address
        self._address_space_size = address_space_size
        self._logger = logger

        if address_endianness not in valid_endianness:
            raise RuntimeError(f"An invalid endianness was set for the device I2C address: {address_endianness}")
        if word_endianness not in valid_endianness:
            raise RuntimeError(f"An invalid endianness was set for the device I2C word address: {word_endianness}")
        if read_type not in valid_read_type:
            raise RuntimeError(f"An invalid read type was chosen: {read_type}")
        if write_type not in valid_write_type:
            raise RuntimeError(f"An invalid write type was chosen: {write_type}")

        self._i2c_connection = i2c_connection
        self._address_bitlength = address_bitlength
        self._address_endianness = address_endianness
        self._word_bitlength = word_bitlength
        self._word_endianness = word_endianness
        self._read_type = read_type
        self._write_type = write_type

        self._not_read = True
        self._memory = [None for _ in range(self._address_space_size)]
        self._defaults = [None for _ in range(self._address_space_size)]
        self._read_only_map = [True for _ in range(self._address_space_size)]
        self._blocks = {}
        self._register_map = {}
        self._register_blocks = {}
        for block_name in address_space_map:
            if "Base Address" in address_space_map[block_name]:
                base_address = address_space_map[block_name]["Base Address"]

                self._blocks[block_name] = {
                    "Base Address": base_address,
                    "Length": len(
                        address_space_map[block_name]["Registers"]
                    ),  # Note: Assuming that all the listed registers in a block are contiguous in the memory space
                }
                if "Write Base Address" in address_space_map[block_name]:
                    self._blocks[block_name]["Write Base Address"] = address_space_map[block_name]["Write Base Address"]

                for register in address_space_map[block_name]["Registers"]:
                    offset = address_space_map[block_name]["Registers"][register]["offset"]
                    default = address_space_map[block_name]["Registers"][register]["default"]
                    full_address = base_address + offset
                    read_only = False
                    if 'read_only' in address_space_map[block_name]["Registers"][register]:
                        read_only = address_space_map[block_name]["Registers"][register]['read_only']

                    self._register_map[block_name + "/" + register] = full_address
                    self._read_only_map[full_address] = read_only
                    self._memory[full_address] = default
                    self._defaults[full_address] = default
            elif "Indexer" in address_space_map[block_name]:
                indexer_info = address_space_map[block_name]['Indexer']
                min_address, max_address, base_addresses = self._get_indexed_block_address_range(
                    block_name, indexer_info, address_space_map[block_name]['Registers']
                )

                if (
                    max_address >= min_address
                ):  # Note: even though not frequently used, a block covering the whole array is needed for bulk read/write operations
                    self._blocks[block_name] = {"Base Address": min_address, "Length": max_address - min_address + 1}

                for block_ref in base_addresses:  # Note: it is a block ref and not a block name because this is a block array
                    self._blocks[block_ref] = {
                        "Base Address": base_addresses[block_ref]['base_address'],
                        "Length": len(
                            address_space_map[block_name]["Registers"]
                        ),  # Note: Assuming that all the listed registers in a block are contiguous in the memory space
                    }

                for register in address_space_map[block_name]["Registers"]:
                    offset = address_space_map[block_name]["Registers"][register]["offset"]
                    default = address_space_map[block_name]["Registers"][register]["default"]
                    for base_name in base_addresses:
                        base_address = base_addresses[base_name]['base_address']
                        full_address = base_address + offset
                        read_only = False
                        if 'read_only' in address_space_map[block_name]["Registers"][register]:
                            read_only = address_space_map[block_name]["Registers"][register]['read_only']

                        full_register_name = base_name + "/" + register
                        self._register_map[full_register_name] = full_address
                        self._read_only_map[full_address] = read_only
                        self._memory[full_address] = default
                        self._defaults[full_address] = default

            else:
                self._logger.error(
                    "An impossible condition occured, there was a memory block defined which does not have a base address and does not have an indexer"
                )

        self._bytes_per_word = ceil(self._word_bitlength / 8)
        self._has_readonly = False
        for val in self._read_only_map:
            if val:
                self._has_readonly = True
                break

    def __len__(self):
        return self._address_space_size

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._memory[index]
        elif isinstance(index, tuple):
            block_name, register_name = index
            return self._memory[self._register_map[block_name + "/" + register_name]]
        else:
            raise RuntimeError("Unknown index for address space controller get")

    def __setitem__(self, index, value):
        if isinstance(index, int):
            self._memory[index] = value
        elif isinstance(index, tuple):
            block_name, register_name = index
            self._memory[self._register_map[block_name + "/" + register_name]] = value
        else:
            raise RuntimeError("Unknown index for address space controller set")

    def __iter__(self):
        for i in self._memory:
            yield i

    def _get_indexed_block_address_range(self, block_name, indexer_info, register_map):
        indexer_function = indexer_info['function']

        block_info = {}
        for idx in range(len(indexer_info['vars'])):
            var = indexer_info['vars'][idx]
            min = indexer_info['min'][idx]
            max = indexer_info['max'][idx]

            old_block_info = block_info
            block_info = {}

            if var == "block" and min is None and max is None:
                param = block_name
                if len(old_block_info) == 0:
                    block_info[param] = {
                        'params': {'block': param},
                    }
                else:
                    for old in old_block_info:
                        block_info[old + ":" + param] = {}
                        block_info[old + ":" + param]['params'] = (old_block_info[old]['params']).copy()
                        block_info[old + ":" + param]['params']['block'] = str(param)
            else:
                for val_idx in range(max - min):
                    i = min + val_idx
                    if len(old_block_info) == 0:
                        block_info[i] = {
                            'params': {var: i},
                        }
                    else:
                        for old in old_block_info:
                            block_info[old + ":" + str(i)] = {}
                            block_info[old + ":" + str(i)]['params'] = (old_block_info[old]['params']).copy()
                            block_info[old + ":" + str(i)]['params'][var] = i

        min_address = None
        max_address = None
        for key in block_info:
            address = indexer_function(**(block_info[key]['params']))
            block_info[key]['base_address'] = address
            if min_address is None or address < min_address:
                min_address = address
            if max_address is None or address > max_address:
                max_address = address

        max_offset = None
        for register in register_map:
            offset = register_map[register]['offset']
            if max_offset is None or offset > max_offset:
                max_offset = offset

        return min_address, max_address + max_offset, block_info

    def update_i2c_address(self, address: int):
        if address != self._i2c_address:
            self._i2c_address = address
            self._not_read = True

            if address is not None:
                self._logger.info(f"Updated address space '{self._name}' to the I2C address {address:#04x}")
            else:
                self._logger.info(f"Reset the I2C address for the address space '{self._name}'")

    # def update_register_map(self, register_map: dict[str, int]):
    #    self._register_map = register_map

    def get_register(self, block_name, register_name):
        return self._memory[self._register_map[block_name + "/" + register_name]]

    def get_register_address(self, block_name, register_name):
        return self._register_map[block_name + "/" + register_name]

    def set_register(self, block_name, register_name, value):
        self._memory[self._register_map[block_name + "/" + register_name]] = value

    def read_memory_block(self, base_address, word_count):
        if self._i2c_address is None:
            self._logger.error(f"Unable to read address space '{self._name}' because the i2c address is not set")
            return False

        self._logger.info(
            f"Reading a block of {word_count} words ({self._bytes_per_word} bytes each) starting at address {base_address} in the address space '{self._name}'"
        )

        tmp = self._i2c_connection.read_device_memory(
            self._i2c_address,
            base_address,
            word_count=word_count,
            address_bitlength=self._address_bitlength,
            address_endianness=self._address_endianness,
            word_bitlength=self._word_bitlength,
            word_endianness=self._word_endianness,
            read_type=self._read_type,
        )
        for idx in range(word_count):
            self._memory[base_address + idx] = tmp[idx]

        return True

    def read_memory_word(self, address):
        return self.read_memory_block(address, 1)

    def write_memory_block(self, base_address, word_count, readback_check: bool = True, readback_base_address=None):
        if self._i2c_address is None:
            self._logger.error(f"Unable to write address space '{self._name}' because the i2c address is not set")
            return False

        has_read_only = False
        for idx in range(word_count):
            if self._read_only_map[base_address + idx]:
                has_read_only = True
                break
        if has_read_only:
            self._logger.info(
                f"The block of {word_count} words starting at address {base_address} in the address space '{self._name}' covers one or more words which are read only, it will be broken down into smaller blocks which do not cover the read only words"
            )
            return self.write_memory_block_with_split_for_read_only(base_address, word_count, readback_check, readback_base_address)

        self._logger.info(f"Writing a block of {word_count} words starting at address {base_address} in the address space '{self._name}'")

        self._i2c_connection.write_device_memory(
            self._i2c_address,
            base_address,
            self._memory[base_address : base_address + word_count],
            address_bitlength=self._address_bitlength,
            address_endianness=self._address_endianness,
            word_bitlength=self._word_bitlength,
            word_endianness=self._word_endianness,
            write_type=self._write_type,
        )

        if readback_check:
            if readback_base_address is None:
                readback_base_address = base_address

            tmp = self._i2c_connection.read_device_memory(
                self._i2c_address,
                readback_base_address,
                word_count=word_count,
                address_bitlength=self._address_bitlength,
                address_endianness=self._address_endianness,
                word_bitlength=self._word_bitlength,
                word_endianness=self._word_endianness,
                read_type=self._read_type,
            )

            failed = []
            for idx in range(word_count):
                if self._memory[base_address + idx] != tmp[idx]:
                    failed += [base_address + idx]
                    self._memory[base_address + idx] = tmp[idx]
            if len(failed) != 0:
                failed = ["0x{:0x}".format(i) for i in failed]
                self._logger.error(
                    f"Failure to write memory block at address {base_address:#0x} with {word_count} words in the '{self._name}' address space (I2C address {self._i2c_address:#02x}). The following addresses failed to write: {', '.join(failed)}"
                )
                return False

        return True

    def write_memory_word(self, address, readback_check: bool = True, readback_address=None):
        if self._read_only_map[address]:
            self._logger.info(
                f"Unable to write to the word at address {address:#0x} in the address space '{self._name}' because it is read only"
            )
            return False

        return self.write_memory_block(address, 1, readback_check, readback_address)

    def write_memory_block_with_split_for_read_only(
        self, base_address, word_count, readback_check: bool = True, readback_base_address=None
    ):
        start_address = None
        ranges = []

        if readback_base_address is None:
            readback_base_address = base_address

        for idx in range(word_count):
            if not self._read_only_map[base_address + idx] and start_address is None:
                start_address = base_address + idx
            if self._read_only_map[base_address + idx] and start_address is not None:
                ranges += [(start_address, base_address + idx - start_address, start_address - base_address + readback_base_address)]
                start_address = None
        if start_address is not None:
            ranges += [(start_address, base_address + idx - start_address + 1, start_address - base_address + readback_base_address)]

        success = True
        self._logger.info(f"Found {len(ranges)} ranges without read only registers")
        for range_param in ranges:
            if not self.write_memory_block(range_param[0], range_param[1], readback_check, range_param[2]):
                success = False

        return success

    def read_all(self):
        if self._i2c_address is None:
            self._logger.error(f"Unable to read address space '{self._name}' because the i2c address is not set")
            return

        self._logger.info(f"Reading the full '{self._name}' address space")

        if self.read_memory_block(0, self._address_space_size):
            self._not_read = False

    def read_block(self, block_name):
        if self._i2c_address is None:
            self._logger.error(f"Unable to read address space '{self._name}' because the i2c address is not set")
            return

        block = self._blocks[block_name]
        self._logger.info("Attempting to read block {}".format(block_name))

        self.read_memory_block(block["Base Address"], block["Length"])

    def read_register(self, block_name, register_name):
        self._logger.detailed_trace(f'Address_Space_Controller::read_register("{block_name}", "{register_name}")')
        if self._i2c_address is None:
            self._logger.error(f"Unable to read address space '{self._name}' because the i2c address is not set")
            return

        self._logger.info(f"Attempting to read register {register_name} in block {block_name}")

        self.read_memory_block(self._register_map[block_name + "/" + register_name], 1)

    def write_all(self, readback_check: bool = True):
        if self._i2c_address is None:
            self._logger.error(f"Unable to write address space '{self._name}' because the i2c address is not set")
            return False

        if self._has_readonly:
            self._logger.info(
                f"Unable to write the full '{self._name}' address space because there are some read only registers, breaking it into smaller blocks"
            )
            return self.write_memory_block_with_split_for_read_only(0, self._address_space_size, readback_check)

        self._logger.info("Writing the full '{}' address space".format(self._name))
        return self.write_memory_block(0, self._address_space_size, readback_check)

    def write_block(self, block_name, readback_check: bool = True):
        if self._i2c_address is None:
            self._logger.error(f"Unable to write address space '{self._name}' because the i2c address is not set")
            return False

        block = self._blocks[block_name]
        self._logger.info("Attempting to write block {}".format(block_name))

        base_address = block["Base Address"]
        original_base_address = base_address
        if "Write Base Address" in block:
            base_address = block["Write Base Address"]

        return self.write_memory_block(base_address, block["Length"], readback_check, original_base_address)

    def write_register(self, block_name, register_name, readback_check: bool = True):
        self._logger.detailed_trace(f'Address_Space_Controller::write_register("{block_name}", "{register_name}", {readback_check})')
        if self._i2c_address is None:
            self._logger.error(f"Unable to write address space '{self._name}' because the i2c address is not set")
            return False

        self._logger.info("Attempting to write register {} in block {}".format(register_name, block_name))

        address = self._register_map[block_name + "/" + register_name]
        original_address = address
        if "Write Base Address" in self._blocks[block_name]:
            new_base = self._blocks[block_name]["Write Base Address"]
            old_base = self._blocks[block_name]["Base Address"]
            address = address - old_base + new_base

        return self.write_memory_block(address, 1, readback_check, original_address)

    def reset_to_defaults(self):
        # register_count = len(self._register_map)
        # count = 0
        for address in self._memory:
            self._memory[address] = self._defaults[address]
            # count += 1



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from __future__ import annotations

from typing import Union

from usb_iss import UsbIss
from usb_iss import defs

from .i2c_connection_helper import I2C_Connection_Helper
from .i2c_messages import I2CMessages

valid_clocks = [20, 50, 100, 400, 1000]
software_clocks = [20, 50, 100, 400]
hardware_clocks = [100, 400, 1000]


class USB_ISS_Helper(I2C_Connection_Helper):
    """Class to handle the USB-ISS connection

    This class is a wrapper around the usb_iss library, providing a simplified interface for the
    usage in the ETROC DAQ assuming only I2C is used, with optional GPIO for the extra pins

    Parameters
    ----------
    port
        The port where the USB-ISS device can be found

    clock
        The clock frequency to use for I2C communication. It must be a valid clock from the
        list: 20, 50, 100, 400, 1000. For clock frequencies where there is both software and
        hardware support, the hardware version will be given precedence.

    use_serial
        Whether to use the extra pins on the USB-ISS as a serial interface, if not they will be
        GPIO (by default set to analog in). The serial port is only enabled if the baud rate is
        also set.

    baud_rate
        The baud rate to use for the serial communication

    verbose
        Whether to output detailed information to the terminal

    max_seq_byte
        The maximum number of sequential bytes supported in a single I2C message. The limit is not
        necessarily from the USB-ISS and may be from the device it is communicating with. TODO: Add
        an option to set this limit per operation so different limits can be set for different devices.

    dummy_connect
        If set, the connection to the USB-ISS will be emulated and a dummy device will be configured

    Raises
    ------
    SerialException
        If the serial object can not be found

    ValueError
        If a wrong value is passed to the clock configuration

    RuntimeError
        For other issues communicating with the USB-ISS device

    Examples
    --------
    >>> import i2c_gui2
    >>> usbiss = i2c_gui2.USB_ISS_Helper("/dev/ttyACM0")
    >>> usbiss.version()

    """

    _port: str
    _clock: int
    _verbose: bool
    _use_serial: bool
    _baud_rate: Union[int, None]
    _iss: UsbIss
    _fw_version: int
    _serial: str

    def __init__(
        self,
        port: str,
        clock: int,
        use_serial: bool = False,
        baud_rate: Union[int, None] = None,
        verbose: bool = False,
        max_seq_byte: int = 8,
        dummy_connect: bool = False,
    ):
        super().__init__(max_seq_byte=max_seq_byte, no_connect=dummy_connect)
        if clock not in valid_clocks:
            raise ValueError(f"Received a wrong clock value: {clock} kHz")

        self._port = port
        self._clock = clock
        self._verbose = verbose
        self._use_serial = use_serial
        self._baud_rate = baud_rate

        if self._use_serial and self._baud_rate is None:
            self._use_serial = False

        if not self._use_serial:
            self._baud_rate = None

        self._iss = UsbIss(dummy=dummy_connect, verbose=self._verbose)
        self._iss.open(port)

        module_id = self._iss.read_module_id()
        if module_id != 7 and not dummy_connect:
            raise RuntimeError(f"Got an unexpected value for the module ID of the USB-ISS device: {module_id}")

        self._fw_version = self._iss.read_fw_version()
        self._serial = self._iss.read_serial_number()

        self._use_hardware = False
        if self._clock in hardware_clocks:
            self._use_hardware = True

        if not self._use_serial:
            self._iss.setup_i2c(self._clock, self._use_hardware, defs.IOType.ANALOGUE_INPUT, defs.IOType.ANALOGUE_INPUT)
        else:
            self._iss.setup_i2c_serial(self._clock, self._use_hardware, self._baud_rate)

        self._is_connected = True

    def __del__(self):
        if hasattr(self, "_iss"):
            if self._iss is not None:
                self._iss.close()

    @property
    def fw_version(self) -> int:
        """The fw_version property getter method

        This method returns the firmware version of the USB-ISS firmware

        Returns
        -------
        int
            The firmware version number.
        """
        return self._fw_version

    @property
    def serial(self) -> str:
        """The serial_number property getter method

        This method returns the serial number of the USB-ISS firmware

        Returns
        -------
        str
            The serial number.
        """
        return self._serial

    @property
    def port(self) -> str:
        """The port property getter method

        This method returns the port the USB-ISS is connected to

        Returns
        -------
        str
            The port.
        """
        return self._port

    @property
    def clock(self) -> int:
        """The clock property getter method

        This method returns the clock frequency the USB-ISS I2C is configured to

        Returns
        -------
        int
            The clock frequency.
        """
        return self._clock

    @property
    def use_serial(self) -> bool:
        """The use_serial property getter method

        This method returns if the extra pins on the USB-ISS are being used as a serial communication

        Returns
        -------
        bool
            If the extra pins on the USB-ISS are being used as a serial communication
        """
        return self._use_serial

    @property
    def baud_rate(self) -> Union[int, None]:
        """The baud_rate property getter method

        This method returns the baud rate the extra pins on the USB-ISS used as a serial communication are set to

        Returns
        -------
        int
            If the extra pins on the USB-ISS are being used as a serial communication
        None
            If the USB-ISS is not configured to use the serial port for communication
        """
        return self._baud_rate

    def _check_i2c_device(self, device_address: int) -> bool:
        """The internal method to check if an i2c device with the given address is connected

        This method overrides the one from the base class.

        Parameters
        ----------
        device_address
            The I2C address of the device to check. It must be a 7-bit address as per the I2C standard

        Returns
        -------
        bool
            The presence or absence of the device with the `device_address`
        """
        return self._iss.i2c.test(device_address)

    def _write_i2c_device_memory(
        self,
        device_address: int,
        word_address: int,
        byte_data: list[int],
        write_type: str = 'Normal',
        address_bitlength: int = 8,
    ):
        """The internal method to write byte data to an i2c device with the given address.

        This method overrides the one from the base class.

        Parameters
        ----------
        device_address
            The I2C address of the device to write to. It must be a 7-bit address as per the I2C standard.

        word_address
            The address of the first byte to be written to. The address should have its endianness correctly
            set such that the MSB of `word_address` is the first byte sent on the I2C lines.

        byte_data
            The byte data to be written to the I2C device. The data is written to the I2C bus in the order
            given, so the endianness of the data must be correctly set, assuming the device contains registers
            larger than 8 bits.

        write_type
            The type of protocol used for the actual writing procedure to the device.

        address_bitlength
            The length in bits of the address

        Raises
        ------
        RuntimeError
            If a not supported configuration is chosen
        """
        if write_type == 'Normal':
            if address_bitlength == 16:
                self._iss.i2c.write_ad2(device_address, word_address, byte_data)
            elif address_bitlength == 8:
                self._iss.i2c.write_ad1(device_address, word_address, byte_data)
            else:
                raise RuntimeError("Unknown bit size trying to be sent")
        else:
            raise RuntimeError("Unknown write type chosen for the USB ISS")

    def _read_i2c_device_memory(
        self,
        device_address: int,
        word_address: int,
        byte_count: int,
        read_type: str = 'Normal',
        address_bitlength: int = 8,
    ) -> list[int]:
        """The internal method to read byte data from an i2c device with the given address.

        This method overrides the one from the base class.

        Parameters
        ----------
        device_address
            The I2C address of the device to write to. It must be a 7-bit address as per the I2C standard.

        word_address
            The address of the first byte to be read from. The address should have its endianness correctly
            set such that the MSB of `word_address` is the first byte sent on the I2C lines.

        byte_count
            The number of bytes to be read from the I2C device. If multybyte registers are to be read, the
            number of bytes should be given, not the number of words.

        read_type
            The type of protocol used for the actual reading procedure from the device. Supported protocols
            are "Normal" and "Repeated Start". The Repeated Start protocol is implemented by the AD5593R chip.

        address_bitlength
            The length in bits of the address

        Raises
        ------
        RuntimeError
            If a not supported configuration is chosen

        Returns
        -------
        list[int]
            The list of bytes in the order presented on the I2C bus. If multibyte registers are being read, then
            the words must be correctly put back together by taking into account the data endianness sent by the
            I2C device.
        """
        if read_type == 'Normal':
            if address_bitlength == 16:
                return self._iss.i2c.read_ad2(device_address, word_address, byte_count)
            if address_bitlength == 8:
                return self._iss.i2c.read_ad1(device_address, word_address, byte_count)
            else:
                raise RuntimeError("Unknown bit size trying to be sent")
        elif read_type == "Repeated Start":
            direct_msg = [defs.I2CDirect.START]

            device_address_byte = device_address << 1
            if address_bitlength == 8:
                direct_msg += [
                    defs.I2CDirect.WRITE2,
                    device_address_byte,
                    word_address & 0xFF,
                ]
            elif address_bitlength == 16:
                direct_msg += [
                    defs.I2CDirect.WRITE3,
                    device_address_byte,
                    (word_address >> 8) & 0xFF,
                    word_address & 0xFF,
                ]
            else:
                raise RuntimeError("Unknown bit size trying to be sent")

            direct_msg += [
                defs.I2CDirect.RESTART,
                defs.I2CDirect.WRITE1,
                device_address_byte | 0x01,
            ]

            if byte_count <= 16:
                if byte_count > 1:
                    direct_msg += [
                        getattr(defs.I2CDirect, f"READ{byte_count-1}"),
                    ]
            else:
                raise RuntimeError("USB ISS does not support a block read of more than 16 bytes")

            direct_msg += [
                defs.I2CDirect.NACK,
                defs.I2CDirect.READ1,
                defs.I2CDirect.STOP,
            ]

            retVal = self._iss.i2c.direct(direct_msg)

            if len(retVal) != byte_count:
                raise RuntimeError("Did not receive the expected number of bytes")
            else:
                return retVal
        else:
            raise RuntimeError("Unknown read type chosen for the USB ISS")

    def _direct_i2c(self, commands: list[I2CMessages]) -> list[int]:  # noqa: C901
        """The internal method to send arbitrary I2C messages to the I2C bus.

        This method overrides the one from the base class.

        Parameters
        ----------
        commands
            The I2C commands to be sent to the I2C bus in the order they are to be sent.

        Raises
        ------
        RuntimeError
            If an unknown command is trying to be sent

        Returns
        -------
        list[int]
            The list of bytes returned to the I2C Bus in the order presented on the I2C bus.
        """
        direct_msg = []

        idx = 0
        while True:
            if idx >= len(commands):
                break

            command = commands[idx]
            if command not in I2CMessages:
                raise RuntimeError("Unknown I2C command")

            direct_msg += [command.value]
            if command == I2CMessages.WRITE1:
                direct_msg += [commands[idx + 1]]
                idx += 2
            elif command == I2CMessages.WRITE2:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                idx += 3
            elif command == I2CMessages.WRITE3:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                idx += 4
            elif command == I2CMessages.WRITE4:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                idx += 5
            elif command == I2CMessages.WRITE5:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                idx += 6
            elif command == I2CMessages.WRITE6:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                idx += 7
            elif command == I2CMessages.WRITE7:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                idx += 8
            elif command == I2CMessages.WRITE8:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                idx += 9
            elif command == I2CMessages.WRITE9:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                idx += 10
            elif command == I2CMessages.WRITE10:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                direct_msg += [commands[idx + 10]]
                idx += 11
            elif command == I2CMessages.WRITE11:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                direct_msg += [commands[idx + 10]]
                direct_msg += [commands[idx + 11]]
                idx += 12
            elif command == I2CMessages.WRITE12:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                direct_msg += [commands[idx + 10]]
                direct_msg += [commands[idx + 11]]
                direct_msg += [commands[idx + 12]]
                idx += 13
            elif command == I2CMessages.WRITE13:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                direct_msg += [commands[idx + 10]]
                direct_msg += [commands[idx + 11]]
                direct_msg += [commands[idx + 12]]
                direct_msg += [commands[idx + 13]]
                idx += 14
            elif command == I2CMessages.WRITE14:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                direct_msg += [commands[idx + 10]]
                direct_msg += [commands[idx + 11]]
                direct_msg += [commands[idx + 12]]
                direct_msg += [commands[idx + 13]]
                direct_msg += [commands[idx + 14]]
                idx += 15
            elif command == I2CMessages.WRITE15:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                direct_msg += [commands[idx + 10]]
                direct_msg += [commands[idx + 11]]
                direct_msg += [commands[idx + 12]]
                direct_msg += [commands[idx + 13]]
                direct_msg += [commands[idx + 14]]
                direct_msg += [commands[idx + 15]]
                idx += 16
            elif command == I2CMessages.WRITE16:
                direct_msg += [commands[idx + 1]]
                direct_msg += [commands[idx + 2]]
                direct_msg += [commands[idx + 3]]
                direct_msg += [commands[idx + 4]]
                direct_msg += [commands[idx + 5]]
                direct_msg += [commands[idx + 6]]
                direct_msg += [commands[idx + 7]]
                direct_msg += [commands[idx + 8]]
                direct_msg += [commands[idx + 9]]
                direct_msg += [commands[idx + 10]]
                direct_msg += [commands[idx + 11]]
                direct_msg += [commands[idx + 12]]
                direct_msg += [commands[idx + 13]]
                direct_msg += [commands[idx + 14]]
                direct_msg += [commands[idx + 15]]
                direct_msg += [commands[idx + 16]]
                idx += 17
            else:
                idx += 1

        return self._iss.i2c.direct(direct_msg)



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

import logging

import pytest
import requests


@pytest.fixture(autouse=True)
def disable_network_calls(monkeypatch):
    def stunted_get():
        raise RuntimeError("Network access not allowed during testing!")

    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: stunted_get())


@pytest.fixture
def logger():
    # logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s:%(name)s:%(message)s')
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s:%(name)s:%(message)s')
    yield logging.getLogger("Test_Logger")



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################


import pytest

from i2c_gui2.functions import address_to_phys
from i2c_gui2.functions import bytes_to_word_list
from i2c_gui2.functions import is_valid_hostname
from i2c_gui2.functions import is_valid_ip
from i2c_gui2.functions import swap_endian_16bit
from i2c_gui2.functions import swap_endian_32bit
from i2c_gui2.functions import valid_i2c_address
from i2c_gui2.functions import validate_hostname
from i2c_gui2.functions import word_list_to_bytes


def test_is_valid_hostname_localhost():
    assert is_valid_hostname("localhost")


def test_is_valid_hostname_google():
    assert is_valid_hostname("google.com")


def test_is_valid_hostname_google_rightdot():
    assert is_valid_hostname("google.com.")


def test_is_valid_hostname_numeric():
    assert not is_valid_hostname("12345.123")


def test_is_valid_hostname_name():
    assert not is_valid_hostname("name")


def test_is_valid_hostname_with_space():
    assert not is_valid_hostname("na me.com")
    assert not is_valid_hostname("name.c om")


def test_is_valid_hostname_too_long():
    assert not is_valid_hostname(
        "This is a very long string which should not be found to be a valid hostname but we should test it anyway just to be sure. The string really has to be very long indeed because the maximum length is 253 characters, so it is not as short as it initially sounds..."
    )


def test_is_valid_ip():
    assert is_valid_ip("192.168.1.0")


def test_is_valid_ip_notIP():
    assert not is_valid_ip("192.168.1.300")
    assert not is_valid_ip("192.168.1")


def test_validate_hostname():
    assert validate_hostname("google.com")
    assert validate_hostname("192.168.1.0")


def test_validate_hostname_notValid():
    assert not validate_hostname("192.188.2")


def test_swap_endian_16bit():
    assert swap_endian_16bit(0x1234) == 0x3412


def test_swap_endian_16bit_truncates():
    assert swap_endian_16bit(0x31234) == 0x3412


def test_swap_endian_32bit():
    assert swap_endian_32bit(0x12345678) == 0x78563412


def test_swap_endian_32bit_truncates():
    assert swap_endian_32bit(0xA12345678) == 0x78563412


def test_valid_i2c_addres_wrong_type():
    assert not valid_i2c_address("hello")


def test_valid_i2c_addres_outside_range():
    assert not valid_i2c_address(0x80)
    assert not valid_i2c_address(0xFF)
    assert not valid_i2c_address(-1)


def test_valid_i2c_addres():
    assert valid_i2c_address(0x21)


def test_address_to_phys_8bit_big():
    assert address_to_phys(0x21, bitlength=8, endianness='big') == 0x21


def test_address_to_phys_8bit_little():
    assert address_to_phys(0x21, bitlength=8, endianness='little') == 0x21


def test_address_to_phys_16bit_big():
    assert address_to_phys(0x2113, bitlength=16, endianness='big') == 0x2113


def test_address_to_phys_16bit_little():
    assert address_to_phys(0x2113, bitlength=16, endianness='little') == 0x1321


def test_address_to_phys_32bit_big():
    assert address_to_phys(0x21763454, bitlength=32, endianness='big') == 0x21763454


def test_address_to_phys_32bit_little():
    assert address_to_phys(0x21763454, bitlength=32, endianness='little') == 0x54347621


def test_address_to_phys_fail():
    with pytest.raises(Exception) as e_info:
        address_to_phys(0x21763454, bitlength=23, endianness='little')
    assert e_info.match(r"^Endian swap not implemented for bit length 23")


def test_word_list_to_bytes_8bit_big():
    assert word_list_to_bytes([0x32, 0x86], bytelength=1, endianness='big') == [0x32, 0x86]


def test_word_list_to_bytes_8bit_little():
    assert word_list_to_bytes([0x32, 0x86], bytelength=1, endianness='little') == [0x32, 0x86]


def test_word_list_to_bytes_16bit_big():
    assert word_list_to_bytes([0x3210, 0x8654], bytelength=2, endianness='big') == [0x32, 0x10, 0x86, 0x54]


def test_word_list_to_bytes_16bit_little():
    assert word_list_to_bytes([0x3210, 0x8654], bytelength=2, endianness='little') == [0x10, 0x32, 0x54, 0x86]


def test_word_list_to_bytes_32bit_big():
    assert word_list_to_bytes([0x32103456, 0x86285031], bytelength=4, endianness='big') == [0x32, 0x10, 0x34, 0x56, 0x86, 0x28, 0x50, 0x31]


def test_word_list_to_bytes_32bit_little():
    assert word_list_to_bytes([0x32103456, 0x86285031], bytelength=4, endianness='little') == [
        0x56,
        0x34,
        0x10,
        0x32,
        0x31,
        0x50,
        0x28,
        0x86,
    ]


def test_bytes_to_word_list_8bit_big():
    assert bytes_to_word_list([0x21, 0x34], bytelength=1, endianness='big') == [0x21, 0x34]


def test_bytes_to_word_list_8bit_little():
    assert bytes_to_word_list([0x21, 0x34], bytelength=1, endianness='little') == [0x21, 0x34]


def test_bytes_to_word_list_16bit_big():
    assert bytes_to_word_list([0x21, 0x34], bytelength=2, endianness='big') == [0x2134]


def test_bytes_to_word_list_16bit_little():
    assert bytes_to_word_list([0x21, 0x34], bytelength=2, endianness='little') == [0x3421]


def test_bytes_to_word_list_32bit_big():
    assert bytes_to_word_list([0x21, 0x34, 0x76, 0x12], bytelength=4, endianness='big') == [0x21347612]


def test_bytes_to_word_list_32bit_little():
    assert bytes_to_word_list([0x21, 0x34, 0x76, 0x12], bytelength=4, endianness='little') == [0x12763421]



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from __future__ import annotations

import logging
from math import ceil

from ..i2c_connection_helper import I2C_Connection_Helper
from ..i2c_messages import I2CMessages
from .address_space_controller import Address_Space_Controller
from .base_chip import Base_Chip

ad5593r_version = "0.0.1"

register_model = {
    "AD5593R": {  # Address Space (i.e. separate I2C memory spaces)
        "Memory Size": 256,
        "Register Bits": 8,
        "Register Length": 16,
        "Endianness": "big",
        "Read Type": 'Repeated Start',
        "Register Blocks": {
            "Config_WR": {
                "Base Address": 0x00,
                "Registers": {
                    "NOP": {
                        "offset": 0x00,
                        "default": 0x0000,
                    },
                    "ADC_SEQ": {
                        "offset": 0x02,
                        "default": 0x0000,
                    },
                    "GEN_CTRL_REG": {
                        "offset": 0x03,
                        "default": 0x0000,
                    },
                    "ADC_CONFIG": {
                        "offset": 0x04,
                        "default": 0x0000,
                    },
                    "DAC_CONFIG": {
                        "offset": 0x05,
                        "default": 0x0000,
                    },
                    "PULLDWN_CONFIG": {
                        "offset": 0x06,
                        "default": 0x00FF,
                    },
                    "LDAC_MODE": {
                        "offset": 0x07,
                        "default": 0x0000,
                    },
                    "GPIO_CONFIG": {
                        "offset": 0x08,
                        "default": 0x0000,
                    },
                    "GPIO_OUTPUT": {
                        "offset": 0x09,
                        "default": 0x0000,
                    },
                    "GPIO_INPUT": {
                        "offset": 0x0A,
                        "default": 0x0000,
                    },
                    "PD_REF_CTRL": {
                        "offset": 0x0B,
                        "default": 0x0000,
                    },
                    "GPIO_OPENDRAIN_CONFIG": {
                        "offset": 0x0C,
                        "default": 0x0000,
                    },
                    "IO_TS_CONFIG": {
                        "offset": 0x0D,
                        "default": 0x0000,
                    },
                    "SW_RESET": {
                        "offset": 0x0F,
                        "default": 0x0000,
                    },
                },
            },
            "Config_RD": {
                "Base Address": 0x70,
                "Write Base Address": 0x00,
                "Registers": {
                    "NOP": {
                        "offset": 0x00,
                        "default": 0x0000,
                    },
                    "ADC_SEQ": {
                        "offset": 0x02,
                        "default": 0x0000,
                    },
                    "GEN_CTRL_REG": {
                        "offset": 0x03,
                        "default": 0x0000,
                    },
                    "ADC_CONFIG": {
                        "offset": 0x04,
                        "default": 0x0000,
                    },
                    "DAC_CONFIG": {
                        "offset": 0x05,
                        "default": 0x0000,
                    },
                    "PULLDWN_CONFIG": {
                        "offset": 0x06,
                        "default": 0x00FF,
                    },
                    "LDAC_MODE": {
                        "offset": 0x07,
                        "default": 0x0000,
                    },
                    "GPIO_CONFIG": {
                        "offset": 0x08,
                        "default": 0x0000,
                    },
                    "GPIO_OUTPUT": {
                        "offset": 0x09,
                        "default": 0x0000,
                    },
                    "GPIO_INPUT": {
                        "offset": 0x0A,
                        "default": 0x0000,
                    },
                    "PD_REF_CTRL": {
                        "offset": 0x0B,
                        "default": 0x0000,
                    },
                    "GPIO_OPENDRAIN_CONFIG": {
                        "offset": 0x0C,
                        "default": 0x0000,
                    },
                    "IO_TS_CONFIG": {
                        "offset": 0x0D,
                        "default": 0x0000,
                    },
                    "SW_RESET": {
                        "offset": 0x0F,
                        "default": 0x0000,
                    },
                },
            },
            "ADC_RD": {
                "Base Address": 0x40,
                "Registers": {
                    "Readings": {
                        "offset": 0x00,
                        "default": 0x0000,
                    },
                },
            },
            "DAC_WR": {
                "Base Address": 0x10,
                "Registers": {
                    "DAC0": {
                        "offset": 0x00,
                        "default": 0x0000,
                    },
                    "DAC1": {
                        "offset": 0x01,
                        "default": 0x0000,
                    },
                    "DAC2": {
                        "offset": 0x02,
                        "default": 0x0000,
                    },
                    "DAC3": {
                        "offset": 0x03,
                        "default": 0x0000,
                    },
                    "DAC4": {
                        "offset": 0x04,
                        "default": 0x0000,
                    },
                    "DAC5": {
                        "offset": 0x05,
                        "default": 0x0000,
                    },
                    "DAC6": {
                        "offset": 0x06,
                        "default": 0x0000,
                    },
                    "DAC7": {
                        "offset": 0x07,
                        "default": 0x0000,
                    },
                },
            },
            "DAC_RD": {
                "Base Address": 0x50,
                "Write Base Address": 0x10,
                "Registers": {
                    "DAC0": {
                        "offset": 0x00,
                        "default": 0x0000,
                    },
                    "DAC1": {
                        "offset": 0x01,
                        "default": 0x0000,
                    },
                    "DAC2": {
                        "offset": 0x02,
                        "default": 0x0000,
                    },
                    "DAC3": {
                        "offset": 0x03,
                        "default": 0x0000,
                    },
                    "DAC4": {
                        "offset": 0x04,
                        "default": 0x0000,
                    },
                    "DAC5": {
                        "offset": 0x05,
                        "default": 0x0000,
                    },
                    "DAC6": {
                        "offset": 0x06,
                        "default": 0x0000,
                    },
                    "DAC7": {
                        "offset": 0x07,
                        "default": 0x0000,
                    },
                },
            },
            "GPIO_RD": {
                "Base Address": 0x60,
                "Registers": {
                    "GPIO": {
                        "offset": 0x00,
                        "default": 0x0000,
                    },
                },
            },
        },
    },
}

register_decoding = {}


class AD5593R_Chip(Base_Chip):
    def __init__(
        self,
        i2c_address,
        i2c_connection: I2C_Connection_Helper,
        logger: logging.Logger,
    ):
        super().__init__(
            chip_name="AD5593R",
            i2c_connection=i2c_connection,
            logger=logger,
            software_version=ad5593r_version,
            register_model=register_model,
            register_decoding=register_decoding,
            # indexer_info=self._indexer_info,
        )

        self._i2c_address = i2c_address

        address_space: Address_Space_Controller = self._address_space["AD5593R"]
        address_space.update_i2c_address(self._i2c_address)

    @property
    def i2c_address(self) -> int:
        return self._i2c_address

    @i2c_address.setter
    def i2c_address(self, address: int):
        self._i2c_address = address

        address_space: Address_Space_Controller = self._address_space["AD5593R"]
        address_space.update_i2c_address(self._i2c_address)

    def read_adc_results(self, adc_sequence: int, num_measurements: int):
        if num_measurements == 0:
            return []

        address_space: Address_Space_Controller = self._address_space["AD5593R"]

        # Select channels to be converted by ADC (Step 2 from the "ADC Operation" in the manual)
        address_space.set_register("Config_RD", "ADC_SEQ", adc_sequence)
        self.write_register("AD5593R", "Config_RD", "ADC_SEQ")

        # Select the ADC for reading (Step 3 from the "ADC Operation" in the manual)
        self._i2c_connection._direct_i2c([I2CMessages.START, I2CMessages.WRITE2, self._i2c_address << 1, 0b01000000, I2CMessages.STOP])

        # Actually read the ADC conversion data (Step 4 from the "ADC Operation" in the manual)
        data = []
        num_repeats = ceil(num_measurements / 8)  # each measurement is 2 bytes and we want to read a max of 16 bytes at a time
        for i in range(num_repeats):
            num_bytes = 16
            if i == num_repeats - 1:
                num_bytes = num_measurements * 2 - i * 16

            commands = [I2CMessages.START, I2CMessages.WRITE1, (self._i2c_address << 1) + 1]

            commands += [getattr(I2CMessages, f'READ{num_bytes - 1}')]

            commands += [I2CMessages.NACK, I2CMessages.READ1, I2CMessages.STOP]

            data += self._i2c_connection._direct_i2c(commands)

        # Stop ADC conversion (explained at the end of the "ADC Operation" in the manual)
        address_space.set_register("Config_RD", "ADC_SEQ", 0x0000)
        self.write_register("AD5593R", "Config_RD", "ADC_SEQ")

        retVal = [None for _ in range(num_measurements)]
        for i in range(num_measurements):
            retVal[i] = ((data[i * 2] & 0xFF) << 8) | (data[i * 2 + 1] & 0xFF)

        return retVal



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################
"""The i2c_messages module

Contains the I2CMessages class, which is only a utility class to help building
custom I2C command sequences using user friendly names.

"""

from __future__ import annotations

from enum import Enum


class I2CMessages(Enum):
    """Class to wrap the I2C Messages unique identifiers

    This is a temporary description

    Examples
    --------
    >>> from i2c_gui2.i2c_messages import I2CMessages
    >>> single_command = I2CMessages.READ7
    >>> command_sequence = [I2CMessages.START, I2CMessages.NACK, I2CMessages.READ2, I2CMessages.STOP]

    """

    # nb: using same IDs as defined for the USB-ISS
    START = 0x01
    RESTART = 0x02
    STOP = 0x03
    NACK = 0x04  # NACK after next read
    READ1 = 0x20
    READ2 = 0x21
    READ3 = 0x22
    READ4 = 0x23
    READ5 = 0x24
    READ6 = 0x25
    READ7 = 0x26
    READ8 = 0x27
    READ9 = 0x28
    READ10 = 0x29
    READ11 = 0x2A
    READ12 = 0x2B
    READ13 = 0x2C
    READ14 = 0x2D
    READ15 = 0x2E
    READ16 = 0x2F
    WRITE1 = 0x30
    WRITE2 = 0x31
    WRITE3 = 0x32
    WRITE4 = 0x33
    WRITE5 = 0x34
    WRITE6 = 0x35
    WRITE7 = 0x36
    WRITE8 = 0x37
    WRITE9 = 0x38
    WRITE10 = 0x39
    WRITE11 = 0x3A
    WRITE12 = 0x3B
    WRITE13 = 0x3C
    WRITE14 = 0x3D
    WRITE15 = 0x3E
    WRITE16 = 0x3F



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from unittest.mock import call
from unittest.mock import patch

import pytest
from usb_iss import defs

from i2c_gui2.i2c_messages import I2CMessages
from i2c_gui2.i2c_usb_iss_helper import USB_ISS_Helper


@pytest.fixture
def port():
    yield "port"


@pytest.fixture
def clock():
    yield 100


@pytest.fixture
def use_serial():
    yield False


@pytest.fixture
def baud_rate():
    yield 96000


@pytest.fixture
def verbose():
    yield True


@pytest.fixture
def max_seq_byte():
    yield 8


@pytest.fixture
def dummy_connect():
    yield False


@pytest.fixture
def usb_iss_test(port, clock, use_serial, baud_rate, verbose, max_seq_byte, dummy_connect):
    yield USB_ISS_Helper(
        port=port,
        clock=clock,
        use_serial=use_serial,
        baud_rate=baud_rate,
        verbose=verbose,
        max_seq_byte=max_seq_byte,
        dummy_connect=dummy_connect,
    )


@pytest.fixture
def mock_usb_iss(port, clock, use_serial, baud_rate, verbose, max_seq_byte, dummy_connect):
    with patch('i2c_gui2.i2c_usb_iss_helper.UsbIss') as mock_class:
        mock_class.return_value = mock_class
        yield mock_class


@pytest.fixture
def usb_iss_mocked(mock_usb_iss, port, clock, use_serial, baud_rate, verbose, max_seq_byte, dummy_connect):
    mock_usb_iss.open.return_value = True
    mock_usb_iss.read_module_id.return_value = 7
    mock_usb_iss.read_fw_version.return_value = "My Version"
    mock_usb_iss.read_serial_number.return_value = "My Serial"

    yield USB_ISS_Helper(
        port=port,
        clock=clock,
        use_serial=use_serial,
        baud_rate=baud_rate,
        verbose=verbose,
        max_seq_byte=max_seq_byte,
        dummy_connect=dummy_connect,
    )

    mock_usb_iss.assert_called_once()
    mock_usb_iss.assert_has_calls([call(dummy=dummy_connect, verbose=verbose)])

    mock_usb_iss.open.assert_called_once()

    mock_usb_iss.read_module_id.assert_called_once()

    mock_usb_iss.read_fw_version.assert_called_once()

    mock_usb_iss.read_serial_number.assert_called_once()

    use_hardware = False
    if clock >= 100:
        use_hardware = True

    if not (use_serial and baud_rate is not None):
        mock_usb_iss.setup_i2c.assert_called_once()
        mock_usb_iss.setup_i2c.assert_has_calls([call(clock, use_hardware, defs.IOType.ANALOGUE_INPUT, defs.IOType.ANALOGUE_INPUT)])
    else:
        mock_usb_iss.setup_i2c_serial.assert_called_once()
        mock_usb_iss.setup_i2c_serial.assert_has_calls([call(clock, use_hardware, baud_rate)])

    # mock_usb_iss.close.assert_not_called()


@pytest.mark.parametrize('use_serial', [True, False])
def test_init(port, clock, use_serial, baud_rate, usb_iss_mocked):
    assert usb_iss_mocked.connected

    assert usb_iss_mocked.fw_version == "My Version"
    assert usb_iss_mocked.serial == "My Serial"
    assert usb_iss_mocked.port == port
    assert usb_iss_mocked.clock == clock
    assert usb_iss_mocked.use_serial == use_serial
    if use_serial:
        assert usb_iss_mocked.baud_rate == baud_rate


@pytest.mark.parametrize('dummy_connect', [True])
def test_no_mock(usb_iss_test):
    assert usb_iss_test.connected


def test_invalid_clock(port, use_serial, baud_rate, verbose, max_seq_byte, dummy_connect):
    with pytest.raises(Exception) as e_info:
        USB_ISS_Helper(
            port=port,
            clock=5000,
            use_serial=use_serial,
            baud_rate=baud_rate,
            verbose=verbose,
            max_seq_byte=max_seq_byte,
            dummy_connect=dummy_connect,
        )
    assert e_info.match(r"^Received a wrong clock value: 5000 kHz")


@pytest.mark.parametrize('use_serial', [True])
@pytest.mark.parametrize('baud_rate', [None])
def test_serial_override(usb_iss_mocked):
    assert not usb_iss_mocked.use_serial


@pytest.mark.parametrize("connected", [True, False])
@pytest.mark.parametrize("device_address", [0x30])
def test__check_i2c_device(mock_usb_iss, usb_iss_mocked, connected, device_address):
    mock_usb_iss.i2c = mock_usb_iss
    mock_usb_iss.test.return_value = connected

    mock_usb_iss.test.assert_not_called()
    assert usb_iss_mocked._check_i2c_device(device_address) == connected


@pytest.mark.parametrize("write_type,bitlength", [("Normal", 8), ("Normal", 16)])
@pytest.mark.parametrize("device_address", [0x30])
@pytest.mark.parametrize("word_address", [0x14])
@pytest.mark.parametrize("data", [[0x01, 0x21], [0x13, 0x34, 0x42]])
def test__write_i2c_device_memory(mock_usb_iss, usb_iss_mocked, write_type, bitlength, device_address, word_address, data):
    mock_usb_iss.i2c = mock_usb_iss
    usb_iss_mocked._write_i2c_device_memory(device_address, word_address, data, write_type, bitlength)

    if write_type == "Normal":
        if bitlength == 16:
            mock_usb_iss.write_ad2.assert_has_calls([call(device_address, word_address, data)])
        elif bitlength == 8:
            mock_usb_iss.write_ad1.assert_has_calls([call(device_address, word_address, data)])


@pytest.mark.parametrize("write_type,bitlength", [("Alternate", 16), ("Normal", 24)])
@pytest.mark.parametrize("device_address", [0x30])
@pytest.mark.parametrize("word_address", [0x14])
@pytest.mark.parametrize("data", [[0x01, 0x21]])
def test_fail__write_i2c_device_memory(mock_usb_iss, usb_iss_mocked, write_type, bitlength, device_address, word_address, data):
    mock_usb_iss.i2c = mock_usb_iss
    with pytest.raises(Exception) as e_info:
        usb_iss_mocked._write_i2c_device_memory(device_address, word_address, data, write_type, bitlength)

    if write_type == "Normal":
        if bitlength not in [8, 16]:
            assert e_info.match(r"^Unknown bit size trying to be sent")
    else:
        assert e_info.match(r"^Unknown write type chosen for the USB ISS")


@pytest.mark.parametrize("read_type,bitlength", [("Normal", 8), ("Normal", 16), ("Repeated Start", 8), ("Repeated Start", 16)])
@pytest.mark.parametrize("device_address", [0x30])
@pytest.mark.parametrize("word_address", [0x14])
@pytest.mark.parametrize("words", [2])
def test__read_i2c_device_memory(mock_usb_iss, usb_iss_mocked, read_type, bitlength, device_address, word_address, words):
    mock_usb_iss.i2c = mock_usb_iss
    data = [i for i in range(words)]
    if read_type == "Normal":
        if bitlength == 16:
            mock_usb_iss.read_ad2.return_value = data
        elif bitlength == 8:
            mock_usb_iss.read_ad1.return_value = data
    elif read_type == "Repeated Start":
        mock_usb_iss.direct.return_value = data

    read_data = usb_iss_mocked._read_i2c_device_memory(device_address, word_address, words, read_type, bitlength)

    if read_type == "Normal":
        if bitlength == 16:
            mock_usb_iss.read_ad2.assert_has_calls([call(device_address, word_address, words)])
            assert read_data == data
        elif bitlength == 8:
            mock_usb_iss.read_ad1.assert_has_calls([call(device_address, word_address, words)])
            assert read_data == data
    elif read_type == "Repeated Start":
        if bitlength == 16:
            direct_msg = [
                defs.I2CDirect.START,
                defs.I2CDirect.WRITE3,
                device_address << 1,
                (word_address >> 8) & 0xFF,
                word_address & 0xFF,
                defs.I2CDirect.RESTART,
                defs.I2CDirect.WRITE1,
                (device_address << 1) | 0x01,
                getattr(defs.I2CDirect, f"READ{words-1}"),
                defs.I2CDirect.NACK,
                defs.I2CDirect.READ1,
                defs.I2CDirect.STOP,
            ]
        if bitlength == 8:
            direct_msg = [
                defs.I2CDirect.START,
                defs.I2CDirect.WRITE2,
                device_address << 1,
                word_address & 0xFF,
                defs.I2CDirect.RESTART,
                defs.I2CDirect.WRITE1,
                (device_address << 1) | 0x01,
                getattr(defs.I2CDirect, f"READ{words-1}"),
                defs.I2CDirect.NACK,
                defs.I2CDirect.READ1,
                defs.I2CDirect.STOP,
            ]
        mock_usb_iss.direct.assert_has_calls([call(direct_msg)])
        assert read_data == data


@pytest.mark.parametrize("read_type,bitlength", [("Normal", 24), ("Repeated Start", 8), ("Repeated Start", 12), ("Alternate", 8)])
@pytest.mark.parametrize("device_address", [0x30])
@pytest.mark.parametrize("word_address", [0x14])
@pytest.mark.parametrize("words", [2, 24])
def test_fail__read_i2c_device_memory(mock_usb_iss, usb_iss_mocked, read_type, bitlength, device_address, word_address, words):
    mock_usb_iss.i2c = mock_usb_iss
    mock_usb_iss.direct.return_value = [i for i in range(words - 1)]
    with pytest.raises(Exception) as e_info:
        usb_iss_mocked._read_i2c_device_memory(device_address, word_address, words, read_type, bitlength)

    if read_type == "Normal":
        if bitlength not in [8, 16]:
            assert e_info.match(r"^Unknown bit size trying to be sent")
    elif read_type == "Repeated Start":
        if bitlength not in [8, 16]:
            assert e_info.match(r"^Unknown bit size trying to be sent")
        else:
            if words > 16:
                assert e_info.match(r"^USB ISS does not support a block read of more than 16 bytes")
            else:
                mock_usb_iss.direct.assert_called_once()
                assert e_info.match(r"^Did not receive the expected number of bytes")
    else:
        assert e_info.match(r"^Unknown read type chosen for the USB ISS")


def test__direct_i2c(mock_usb_iss, usb_iss_mocked):
    mock_usb_iss.i2c = mock_usb_iss
    mock_usb_iss.direct.return_value = "My very unique string"

    commands = [
        I2CMessages.NACK,
        I2CMessages.WRITE1,
        2,
        I2CMessages.WRITE2,
        2,
        2,
        I2CMessages.WRITE3,
        2,
        2,
        2,
        I2CMessages.WRITE4,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE5,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE6,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE7,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE8,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE9,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE10,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE11,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE12,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE13,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE14,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE15,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        I2CMessages.WRITE16,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
    ]
    message = usb_iss_mocked._direct_i2c(commands)

    mock_usb_iss.direct.assert_called_once()
    assert message == "My very unique string"


def test_fail__direct_i2c(mock_usb_iss, usb_iss_mocked):
    mock_usb_iss.i2c = mock_usb_iss

    commands = [256]

    with pytest.raises(Exception) as e_info:
        usb_iss_mocked._direct_i2c(commands)

    assert e_info.match(r"^Unknown I2C command")



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from __future__ import annotations

import logging
import tkinter as tk
import tkinter.font
import tkinter.ttk as ttk  # For themed widgets (gives a more native visual to the elements)

from .base_gui import Base_GUI


class About_Helper:
    _parent: Base_GUI

    def __init__(
        self,
        parent: Base_GUI,
        logger: logging.Logger,
    ):
        self._parent = parent
        self._logger = logger

    def display_about(self):
        if hasattr(self, "_about_window"):
            self._logger.info("About window already open")
            return

        self._about_window = tk.Toplevel(self._parent._root)
        self._about_window.protocol('WM_DELETE_WINDOW', self.close_about_window)

        self._about_window.title("About " + self._parent._title)
        self._about_window.geometry("500x500")
        self._about_window.resizable(0, 0)
        self._about_window.columnconfigure(100, weight=1)
        self._about_window.rowconfigure(100, weight=1)

        self._frame = ttk.Frame(self._about_window, padding="5 5 5 5")
        self._frame.grid(column=100, row=100, sticky=(tk.N, tk.W, tk.E, tk.S))
        self._frame.columnconfigure(100, weight=1)

        self._about_title = tk.Label(self._frame, text=self._parent._title, font=('arial', 20, tkinter.font.BOLD))
        self._about_title.grid(column=100, row=100, sticky='')

        self._parent._about_contents(self._frame, column=100, row=200)

        from .. import __version__

        self._about_info_label = tk.Label(self._frame, text="I2C GUI Library - v{}".format(__version__))
        self._about_info_label.grid(column=100, row=2000, sticky='', pady=(20, 0))

        self._about_copy_label = tk.Label(
            self._frame, justify=tk.LEFT, wraplength=490, text="Tool written and developed by Cristóvão Beirão da Cruz e Silva - © 2024"
        )
        self._about_copy_label.grid(column=100, row=2100, sticky=tk.S)

    def close_about_window(self):
        if not hasattr(self, "_about_window"):
            self._logger.info("About window does not exist")
            return

        self._about_window.destroy()
        del self._about_window



# -*- coding: utf-8 -*-
#############################################################################
# zlib License
#
# (C) 2024 Cristóvão Beirão da Cruz e Silva <cbeiraod@cern.ch>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
#############################################################################

from i2c_gui2.i2c_messages import I2CMessages


def test_start():
    assert I2CMessages.START.value == 0x01


def test_restart():
    assert I2CMessages.RESTART.value == 0x02


def test_stop():
    assert I2CMessages.STOP.value == 0x03


def test_nack():
    assert I2CMessages.NACK.value == 0x04  # NACK after next read


def test_read1():
    assert I2CMessages.READ1.value == 0x20


def test_read2():
    assert I2CMessages.READ2.value == 0x21


def test_read3():
    assert I2CMessages.READ3.value == 0x22


def test_read4():
    assert I2CMessages.READ4.value == 0x23


def test_read5():
    assert I2CMessages.READ5.value == 0x24


def test_read6():
    assert I2CMessages.READ6.value == 0x25


def test_read7():
    assert I2CMessages.READ7.value == 0x26


def test_read8():
    assert I2CMessages.READ8.value == 0x27


def test_read9():
    assert I2CMessages.READ9.value == 0x28


def test_read10():
    assert I2CMessages.READ10.value == 0x29


def test_read11():
    assert I2CMessages.READ11.value == 0x2A


def test_read12():
    assert I2CMessages.READ12.value == 0x2B


def test_read13():
    assert I2CMessages.READ13.value == 0x2C


def test_read14():
    assert I2CMessages.READ14.value == 0x2D


def test_read15():
    assert I2CMessages.READ15.value == 0x2E


def test_read16():
    assert I2CMessages.READ16.value == 0x2F


def test_write1():
    assert I2CMessages.WRITE1.value == 0x30


def test_write2():
    assert I2CMessages.WRITE2.value == 0x31


def test_write3():
    assert I2CMessages.WRITE3.value == 0x32


def test_write4():
    assert I2CMessages.WRITE4.value == 0x33


def test_write5():
    assert I2CMessages.WRITE5.value == 0x34


def test_write6():
    assert I2CMessages.WRITE6.value == 0x35


def test_write7():
    assert I2CMessages.WRITE7.value == 0x36


def test_write8():
    assert I2CMessages.WRITE8.value == 0x37


def test_write9():
    assert I2CMessages.WRITE9.value == 0x38


def test_write10():
    assert I2CMessages.WRITE10.value == 0x39


def test_write11():
    assert I2CMessages.WRITE11.value == 0x3A


def test_write12():
    assert I2CMessages.WRITE12.value == 0x3B


def test_write13():
    assert I2CMessages.WRITE13.value == 0x3C


def test_write14():
    assert I2CMessages.WRITE14.value == 0x3D


def test_write15():
    assert I2CMessages.WRITE15.value == 0x3E


def test_write16():
    assert I2CMessages.WRITE16.value == 0x3F
